# From Luna's https://github.com/LunaSare/opentree-ipynb/blob/master/py/get_subtree_for_rank.py
import os
import re
import sys
import gzip
import shutil
import copy
from opentree import OT


_DEBUG = 1
def debug(msg):
    """short debugging command
    """
    if _DEBUG == 1:
        print(msg)

def download_taxonomy_file(version, loc):
    # download ott taxonomy
    version = str(version)
    path = loc + '/' +'ott' + version
    if os.path.exists(path + '/' + 'taxonomy.tsv'):
        sys.stdout.write("Taxonomy already available at {}".format(path))
        return path
    if os.path.exists(path + '.tgz') == False: #Download the file if you don't have it
        os.system('wget "http://files.opentreeoflife.org/ott/ott{}/ott{}.tgz" -P {}'.format(version, version, loc))
    os.system("tar -xzvf {}/ott3.2.tgz -C {}".format(loc, loc))
    assert os.path.exists(path + '/' + 'taxonomy.tsv')
    return path


def get_forwards_dict(forwards_file):
    """Returns a dictionary with new ott_ids for forwarded ott_ids
    """
    fwd_dict = {}
    fi=open(forwards_file)
    header = fi.readline()
    for lin in fi:
        lii = lin.split()
        fwd_dict[int(lii[0])]=lii[1]
    return fwd_dict

def clean_taxonomy_file(taxonomy_file):
    """Generates a pruned taxonomy.
    cleans up the word 'species' and the flag 'no rank - terminal', which is not associated to higher taxonomic ranks
    """
    taxon_dir = os.path.dirname(taxonomy_file)
    output_path = "{}/taxonomy_clean.tsv".format(taxon_dir)
    if not os.path.exists(output_path):
        # clean taxonomy file, writes cleaned file to taxonomy_clean.tsv
        os.system('grep -a -v "major_rank_conflict" ' + taxonomy_file + ' | egrep -a -v "sibling_higher" | egrep -a -v "varietas" | egrep -a -v "no rank" | egrep -a -v "Incertae" | egrep -a -v "incertae" | egrep -a -v "uncultured" | egrep -a -v "barren" | egrep -a -v "extinct" | egrep -a -v "unplaced" | egrep -a -v "hidden" | egrep -a -v "inconsistent" | egrep -a -v "synonym" > {}'.format(output_path))
    assert os.path.exists(output_path)
    return output_path


def get_ott_ids_for_rank(rank, taxonomy_file, synth_only = True):
    """Returns all the ott_ids for a given rank.
    Args
    rank: (must be in ['species', 'genus', 'family', 'order', 'class'])
    taxonomy_file: path to taxonomy.tsv

    If synth_only == True, will return only ids included in synth. 
    (Does not assess if taxa actaully appear as monophyletic in synth, e.g. if  taxa are broken.)
    """
    assert rank in ['species', 'genus', 'family', 'order', 'class']
    assert os.path.exists(taxonomy_file)
    taxon_dir = os.path.dirname(taxonomy_file)
    output_path = "{}/{}.tsv".format(taxon_dir, rank)
    #if not os.path.exists(output_path):
    os.system("""cat {tf} | awk '$7 == "{r}"' > {op}""".format(tf=taxonomy_file, r=rank, op=output_path))
        # clean taxonomy file
#        os.system('grep -a "' + rank + '" ' + taxonomy_file + ' | egrep -v "Incertae" | egrep -v "no rank" | egrep -v "major_rank_conflict" | egrep -v "uncultured" | egrep -v "barren" | egrep -v "extinct" | egrep -v "incertae" | egrep -v "unplaced" | egrep -v "hidden" | egrep -v "inconsistent"  | egrep -v "synonym" | egrep -v "in ' + rank + '" | egrep -v "species" | egrep -v "genus" | egrep -v "super' + rank + '" | egrep -v "sub' + rank + '" > {}'.format(output_path))
    # extract ott ids from taxonomy reduced file
    fi = open(output_path).readlines()
    ott_ids = []
    for lin in fi:
        lii = lin.split('\t')
        ott_ids.append(lii[0])
    if synth_only == True:
        nodes = ['ott' + idn for idn in ott_ids]
        resp = OT.synth_node_info(node_ids = nodes)
        if 'unknown' in resp.response_dict:
            synth_ids = set(nodes).difference(set(resp.response_dict['unknown']))
            ott_ids = [nodeid.strip('ott') for nodeid in synth_ids]
    return ott_ids


def get_ott_ids_for_group(group_ott_id, write_file = 'children_ott_ids.txt', synth_only = False):
    """Returns all descendent ottids of a taxon"""
    sys.stdout.write('Gathering ott ids from group with ott id {}.\n'.format(group_ott_id))
    #debug(group_ott_id)
    subtree = OT.taxon_subtree(ott_id = group_ott_id, label_format='name_and_id')
    if synth_only == True:
        nodes = [taxon.label.split()[-1] for taxon in subtree.tree.taxon_namespace]
        resp = OT.synth_node_info(node_ids = nodes)
        if 'unknown' in resp.response_dict:
            synth_ids = set(nodes).difference(set(resp.response_dict['unknown']))
            ott_ids = [nodeid.strip('ott') for nodeid in synth_ids]
    else:
        ott_ids =[taxon.label.split()[-1].strip('ott') for taxon in subtree.tree.taxon_namespace]
    return ott_ids


def get_ott_ids_group_and_rank(group_ott_id, rank, taxonomy_file, synth_only=True):
    """Get all ott_ids of rank 'rank' in group 'group_ott_id'
    e.g. get all genera in Aves

    """
    # clean taxonomy file
    # get group ott ids
    children_ott_ids = get_ott_ids_for_group(group_ott_id)
    # get rank ott ids
    rank_ott_ids = get_ott_ids_for_rank(rank, taxonomy_file, synth_only = synth_only)
    # get rank ott ids that are in children ott ids:
    ott_ids = []
    for item in children_ott_ids:
        if item in rank_ott_ids:
            ott_ids.append(item)
    return(ott_ids)


def standardize_labels(tree, prob_char = "():#", replace_w = '_'):
    """While parens in labels is acceptable Newick, some tree viewers (e.g. itol) cannot deal
    This takes a tree and removes troublsesome characters"""
    for taxon in tree.taxon_namespace:
        taxon.label = remove_problem_characters(taxon.label, prob_char, replace_w)
    for node in tree:
        if node.label:
            node.label = remove_problem_characters(node.label, prob_char, replace_w)
    return tree

def remove_problem_characters(instr, prob_char = "():#", replace_w = ''):
    """While colons, parens, etc are leagal in newick labels, 
    many tree viewers won't read or misread trees with them.
    This function replaces problem characters in a string with a replacement char.

    Args:
        instr: The input string
        prob_char: String containing problematic characters. Default  is"():#"
        replace_w: Character to replace problem characters with. Default  is "_"

    Returns:
        String with problem characters replaced."""

    problem_characters = set(prob_char)
    for char in problem_characters:
        instr = instr.replace(char,replace_w)
    return instr

def _gather_broken_taxa_info(broken_response, label_format):
    """Gathers information about broken taxa in a synth tree call

    Args:
        broken_response: induced_subtree_response['broken']
        label format: one of ['name', 'id', 'name_and_id']

    Returns: (relabel, relabel_ott_ids, broken_dict)
        where 
        relabel: {nodeid : [label_broken_tax1, label_broken_tax2]
        relabel_ott_ids: {nodeid : [ottid_broken_tax1, ottid_broken_tax2]
        broken_dict: {ottid_broken_tax1: {'url:  url_to_taxonomy,
                                          'MRCA_location_in_synth': nodeid1,
                                          'broken_taxa_mapping_to_same_node': ottid_broken_tax2,
                                          and all the other responses from OT.taxon_info, e.g.
                                          'name': ott taxon name,
                                          }
                    }
        
    """
    broken_dict = {}
    relabel = {}
    relabel_ott_ids = {}
    for taxon in broken_response:
        remap = broken_response[taxon] # Where on the tree is that taxon now?
        ott_id = taxon.strip('ott')
        tax_inf = OT.taxon_info(ott_id=ott_id).response_dict
        tax_inf['tax_url'] = "https://tree.opentreeoflife.org/taxonomy/browse?id={}".format(ott_id)
        tax_inf['synth_url'] = "https://tree.opentreeoflife.org/opentree/argus/ottol@{}".format(ott_id)
        tax_inf['MRCA_location_in_synth'] = remap
        taxon_name = tax_inf.get('name', taxon)
        if label_format == 'name':
            taxon_label = "{}".format(taxon_name)
        elif label_format == 'name_and_id':
            taxon_label = "{}_{}".format(taxon_name, taxon)
        else:
            taxon_label = taxon
        if remap not in relabel:
            relabel[remap] = [] #Sometimes multiple taxa map to the same node or id
            relabel_ott_ids[remap] = []
        relabel[remap].append("{}".format(taxon_label))
        relabel_ott_ids[remap].append(ott_id)
        tax_inf['broken_taxa_mapping_to_same_node'] = relabel[remap]
        broken_dict[ott_id] = tax_inf
    return relabel, relabel_ott_ids, broken_dict


def _gather_unknown_taxa_info(unknown_ids):
    """Gathers taxon info for unknwon ids

    Args:
        unknown_ids: a list of unknown ott ids

    Returns a dictionary containing:
        unknown_dict: a dictionary with ott_ids as keys
            value: dictionary containnig
                full response of a taxon_info call
                + 'url': the link to the taxon in taxonomy browser

    """
    unknown_dict = {}
    for unk in unknown_ids:
            uid = unk.strip('ott') #URL for taxonomy needs integer
            tax_inf = OT.taxon_info(ott_id=uid).response_dict
            tax_inf['url'] = "https://tree.opentreeoflife.org/taxonomy/browse?id={}".format(uid)
            unknown_dict[unk] = tax_inf
    return unknown_dict

def labelled_induced_synth(ott_ids, label_format = 'name', inc_unlabelled_mrca=False, standardize=True):
    """Interpreting node ids from a search on taxa can be challenging.
    This relabels MRCA based tips with what broken taxa they were replacing.
    Sometimes several query taxa map to the same synth node.

    Args:
        ott_ids: The search taxa, as ott_ids. Can be integers or strings prefixed with "ott"
        label_format: One of 'name', 'id', 'name_and_id'. Default  is "name"
        inc_unlabelled_mrca: Whether to include node labels form synth that do not correspond to taxa
                             Default  is False
        standardize: Whether to standardize taxon labels by removing problem characters. Default is True

    Returns a dictionary containing:
        tree, unknown_ids
        tree: a dendropy tree object with taxa labelled acrording to 'label_format'
        unknown_ids: a list of the input ott_ids that were not found

    """
    
    # call synth tree
    curr_ids = copy.deepcopy(ott_ids) #track who we lost
    call_record = OT.ws.tree_of_life_induced_subtree(ott_ids=curr_ids,
                                                     label_format='name_and_id')#This needs to be id
    # often some ids are not found - cull but track
    if call_record.response_dict.get('unknown'):
        unknown_dict = _gather_unknown_taxa_info(call_record.response_dict['unknown'].keys())
        OT._cull_unknown_ids_from_args(call_record, node_ids=None, ott_ids=curr_ids)
        call_record = OT.ws.tree_of_life_induced_subtree(ott_ids=curr_ids,
                                                     label_format='name_and_id')
    else:
        unknown_dict = {}

    
    assert label_format in ['name', 'id', 'name_and_id'] ##this only apply's to the re-labeled output, not the first call
    relabel, relabel_ott_ids, broken_dict = _gather_broken_taxa_info(call_record.response_dict['broken'], label_format)

    labelled_tree = copy.deepcopy(call_record.tree)
    label_matches = {}
    for taxon in labelled_tree.taxon_namespace:#Are these necessarily tips???
        if taxon.label.startswith('mrca'):
            orig = taxon.label
            assert taxon.label in relabel
            for brok_taxon in relabel_ott_ids[taxon.label]:
                broken_dict[brok_taxon]['is_tip'] = True
            taxon.label = 'MRCA of taxa in '+' '.join(relabel[taxon.label])
            label_matches[orig] = taxon.label
        else:
            orig = taxon.label
            ott_taxon_name = " ".join(taxon.label.split()[:-1])
            ott_taxon_name = remove_problem_characters(ott_taxon_name)
            ott = taxon.label.split()[-1]
            if label_format == 'name':
                new_label = ott_taxon_name
            elif label_format == 'name_and_id':
                new_label = taxon.label
            else:
                new_label = ott
            if ott in relabel:
                added_taxa = ' and MRCA of taxa in '+' '.join(relabel[ott])
                taxon.label = new_label + added_taxa
                for brok_taxon in relabel_ott_ids[ott]:
                    broken_dict[brok_taxon]['is_tip'] = False
            else:
                taxon.label =  new_label
            label_matches[orig] = taxon.label

    for node in labelled_tree:
        if node.label and node.label != '':
            orig = node.label
            if node.taxon:
                node.label = None
            else:
                if node.label.startswith('mrca'):
                    if node.label in relabel:
                        for brok_taxon in relabel_ott_ids[node.label]:
                            broken_dict[brok_taxon]['is_tip'] = False
                        node.label = 'MRCA of taxa in '+' '.join(relabel[node.label])
                        label_matches[orig] = node.label
                    elif inc_unlabelled_mrca:
                        label_matches[orig] = node.label
                        pass
                    else:
                        label_matches[orig] = node.label
                        node.label = None
                         
                else:
                    ott_taxon_name = " ".join(node.label.split()[:-1])
                    ott_taxon_name = remove_problem_characters(ott_taxon_name)
                    ott = node.label.split()[-1]
                    if label_format == 'name':
                        new_label = ott_taxon_name
                    elif label_format == 'name_and_id':
                        new_label = node.label
                    else:
                        new_label = ott
                    if ott in relabel:
                        added_taxa = 'MRCA of taxa in '+' '.join(relabel[ott])
                        node.label = new_label + added_taxa
                        for brok_taxon in relabel_ott_ids[ott]:
                            broken_dict[brok_taxon]['is_tip'] = False
                    else:
                        node.label =  new_label
                    label_matches[orig] = node.label

    if standardize == True:
        labelled_tree = standardize_labels(labelled_tree)
    return {'labelled_tree':labelled_tree,
            'original_tree':call_record.tree,
            'unknown_ids':unknown_dict,
            'non-monophyletic_taxa':broken_dict,
            'supporting_studies': call_record.response_dict["supporting_studies"],
            'label_map':label_matches}


#!/usr/bin/env python3

from .ot_object import OpenTree
from .ws_wrapper import (WebServiceRunMode,
                         )
from . import __version__
import logging
import argparse
import atexit
import sys
import os


def _write_calls_as_curl(ws_wrapper_obj, out=sys.stderr):
    for line in ws_wrapper_obj.curl_strings:
        out.write('{}\n'.format(line))


def process_ott_or_node_id_arg(args):
    ott_id, node_id = None, None
    if args.ott_id:
        unaltered_el = args.ott_id.strip()
        el = unaltered_el[3:] if unaltered_el.startswith('ott') else unaltered_el
        try:
            ott_id = int(el)
        except:
            sys.exit('Expecting each ott ID to be an integer or a string starting with "ott". '
                     'Found "{}"\n'.format(unaltered_el))
    try:
        nia = args.node_id
    except AttributeError:
        nia = None
    if nia:
        node_id = str(nia).strip()
    if node_id and ott_id:
        sys.exit('Expecting either ott-id or node-id, but not both\n')
    return ott_id, node_id


def process_ott_and_node_id_list_args(args):
    ott_id_list, node_id_list = [], []
    if args.ott_ids:
        x = [i.strip().lower() for i in args.ott_ids.split(',')]
        for el in x:
            unaltered_el = el
            if el.startswith('ott'):
                el = el[3:]
            try:
                ott_id_list.append(int(el))
            except:
                sys.exit('Expecting each ott ID to be an integer or a string starting with "ott". '
                         'Found "{}"\n'.format(unaltered_el))
    try:
        # noinspection PyUnusedLocal
        nia = args.node_ids
    except AttributeError:
        pass
    else:
        if args.node_ids:
            node_id_list = [i.strip().lower() for i in args.node_ids.split(',')]
    return ott_id_list, node_id_list


class OTCommandLineTool(object):
    """Helper class for writing a script that uses a common set of Open Tree command line
    options.
    """

    def __init__(self, usage, name=None, common_args=None):
        script_path = 'unknown' if not sys.argv else sys.argv[0]
        if name is None:
            name = os.path.split(script_path)[-1]
        self.name = name
        self.usage = usage
        self.parser = argparse.ArgumentParser(usage)
        self.api_endpoint = None
        self._add_default_open_tree_arguments(common_args=common_args)
        self.ot_factory = None

    def _add_default_open_tree_arguments(self, common_args=None):

        """Adds several standard command line arguments to the command line parser"""
        cli = self.parser
        cli.add_argument('--version', action='store_true', help='request version information and exit')
        cli.add_argument('--logging-level', default='info', type=str,
                         help='sets the logging level. Should be one of: '
                              '"debug", "info", "warning", "error", or "critical"')
        cli.add_argument('--api-endpoint', default="production",
                         help='Advanced option: specifies which server to contact of api calls.'
                              'choices are "production", "dev", "local", "ot" + # or an IP address.')
        cli.add_argument('--run-mode', default='run', type=str,
                         help='Sets the action to take when interacting with the Open Tree API. '
                              '"run" is the normal mode. '
                              '"curl" will emit a curl call to standard error instead of performing the '
                              'web-service call; this usually causes the script to terminate in an error, but'
                              ' the curl call can be helpful for debugging. "curl-on-exit" will perform the'
                              ' web-service calls, and then write the curl calls used to stderr on exit.')
        if common_args and "ott-ids" in common_args:
            cli.add_argument('--ott-ids', default=None, type=str,
                             help='a comma separated list of OTT ids')
        if common_args and "node-ids" in common_args:
            cli.add_argument('--node-ids', default=None, type=str,
                             help='a comma separated list of node ids')
        if common_args and "ott-id" in common_args:
            cli.add_argument('--ott-id', default=None, type=str,
                             help='An OTT ids (integer or string starting with ott')
        if common_args and "node-id" in common_args:
            cli.add_argument('--node-id', default=None, type=str,
                             help='A node id (starting with "mrca" or "ott)')

    def parse_cli(self, arg_list=None):
        """Parses `arg_list` or sys.argv (if None), handles basic options, returns OpenTree and args.

        May call sys.exit - if the user requested an option like --version to display info and exit.

        Returns an OpenTree instance configured with the specified api_endpoint and the args
            object returned by the argparse object's parse_args method"""
        if arg_list is None:
            arg_list = sys.argv[1:]
        args = self.parser.parse_args(arg_list)
        console = logging.StreamHandler()
        logging_level_dict = {"debug": logging.DEBUG,
                              "info": logging.INFO,
                              "warning": logging.WARNING,
                              "error": logging.ERROR,
                              "critical": logging.CRITICAL,
                              }
        log_lev = logging_level_dict.get(args.logging_level.lower())
        if log_lev is None:
            logging.critical('--logging-level="{}" not understood'.format(args.logging_level))
            sys.exit(1)
        console.setLevel(log_lev)
        if args.version:
            m = '{} using opentree python lib version {}\n'
            sys.stderr.write(m.format(self.name, __version__))
            sys.exit(0)
        api_endpoint = args.api_endpoint.lower()
        self.api_endpoint = api_endpoint
        run_mode_dict = {"run": WebServiceRunMode.RUN,
                         "curl": WebServiceRunMode.CURL,
                         "curl-on-exit": WebServiceRunMode.CURL_ON_EXIT
                         }
        run_mode = run_mode_dict.get(args.run_mode.lower())
        if run_mode is None:
            logging.critical('--run-mode="{}" not understood'.format(args.run_mode))
            sys.exit(1)
        self.ot_factory = lambda: OpenTree(api_endpoint, run_mode)
        ot = self.ot_factory()
        if run_mode == WebServiceRunMode.CURL_ON_EXIT:
            atexit.register(_write_calls_as_curl, ot.ws)
        return ot, args



"""OT object. High level wrapper for OpenTree calls"""
#!/usr/bin/env python3
import sys

from .ws_wrapper import (OTWebServicesError,
                         WebServiceRunMode,
                         )
from .ot_ws_wrapper import OTWebServiceWrapper

# from .nexson_helpers import extract_tree_nexson, extract_otu_nexson, detect_nexson_version

FILES_SERVER_URL = 'files'


class FilesServerWrapper(OTWebServiceWrapper):
    """
    This class provides a mid-level wrapper for interaction with OT web services and data.
    """
    def __init__(self, api_endpoint=FILES_SERVER_URL, run_mode=WebServiceRunMode.RUN):
        super(FilesServerWrapper, self).__init__(api_endpoint=api_endpoint, run_mode=run_mode)

    def get_subproblem_scaffold_tree(self, synth_id):
        url_frag = 'synthesis/{s}/{s}/subproblems/subproblems-scaffold-only.tre'.format(s=synth_id)
        return self._call_api(url_frag, http_method='GET', headers='text')

    def get_subproblem_size_info(self, synth_id):
        url_frag = 'synthesis/{s}/{s}/subproblems/subproblem_size_summary.json'.format(s=synth_id)
        return self._call_api(url_frag, http_method='GET')

    def get_subproblem_solution(self, synth_id, ott_id):
        url_frag = 'synthesis/{s}/{s}/subproblem_solutions/ott{o}.tre'.format(s=synth_id,
                                                                              o=ott_id)
        return self._call_api(url_frag, http_method='GET', headers='text')

    def get_reversed_subproblem_solution(self, synth_id, ott_id):
        url_frag = 'synthesis/{s}/{s}/reversed_subproblem_solutions/ott{o}.tre'.format(s=synth_id,
                                                                                       o=ott_id)
        return self._call_api(url_frag, http_method='GET', headers='text')

    def get_subproblem_trees(self, synth_id, ott_id):
        url_frag = 'synthesis/{s}/{s}/subproblems/ott{o}.tre'.format(s=synth_id, o=ott_id)
        return self._call_api(url_frag, http_method='GET', headers='text')


_default_api_endpoint = None
_default_run_mode = None


def default_open_tree_obj():
    global _default_api_endpoint, _default_run_mode
    if _default_api_endpoint is None:
        _default_api_endpoint = 'production'
        _default_run_mode = WebServiceRunMode.RUN
    return OpenTree(api_endpoint=_default_api_endpoint,
                    run_mode=_default_run_mode)


class OpenTree(object):
    """
    This class provides a high-level wrapper for interaction with OT web services and data.
    The method names are intended to be clear to a wide variety of users, rather than
    necessarily matching the API calls directly.
    """

    def __init__(self, api_endpoint='production', run_mode=WebServiceRunMode.RUN):
        global _default_api_endpoint, _default_run_mode
        if _default_api_endpoint is None:
            _default_api_endpoint = api_endpoint
            _default_run_mode = run_mode
        self._api_endpoint = api_endpoint
        self._run_mode = run_mode
        self._ws = None
        self._files_server = None

    @property
    def files_server(self):
        if self._files_server is None:
            self._files_server = FilesServerWrapper(run_mode=self._run_mode)
        return self._files_server

    @property
    def ws(self):
        if self._ws is None:
            self._ws = OTWebServiceWrapper(api_endpoint=self._api_endpoint,
                                           run_mode=self._run_mode)
        return self._ws

    def get_subproblem_scaffold_tree(self, synth_id):
        return self.files_server.get_subproblem_scaffold_tree(synth_id)

    def get_subproblem_size_info(self, synth_id):
        return self.files_server.get_subproblem_size_info(synth_id)

    def get_subproblem_solution(self, synth_id, ott_id):
        return self.files_server.get_subproblem_solution(synth_id, ott_id)

    def get_subproblem_trees(self, synth_id, ott_id):
        return self.files_server.get_subproblem_trees(synth_id, ott_id)

    def get_reversed_subproblem_solution(self, synth_id, ott_id):
        return self.files_server.get_reversed_subproblem_solution(synth_id, ott_id)

    def about(self):
        """
        Get information about the Open Tree of Life taxonomy and the synthetic tree.
        """
        tax_about = self.ws.taxonomy_about()
        tree_about = self.ws.tree_of_life_about()
        return {'taxonomy_about': tax_about.response_dict,
                'synth_tree_about': tree_about.response_dict
                }

    def get_study(self, study_id):
        """
        Get a study and its associated metadata.

        Parameters
        ----------
        study_id : single character value
            The study id from Open Tree of Life.
        """
        return self.ws.study(study_id)

    def get_tree(self, study_id, tree_id,
                 tree_format="nexson", label_format="ot:originallabel", demand_success=False):
        """
        Get a source tree from phylesystem and its associated metadata.

        Parameters
        ----------
        study_id : single character value
            The study id from Open Tree of Life.
        tree_id : single character value
            The tree id of a tree within the study id provided.
        tree_format : single character value
            Must be one of "newick", "nexson", "nexus", or "object"
            If tree format is newick or nexus, returns tree as string in that format.
            If "nexson", returns semi-useless tree nexson w/o OTUS.
        label_format : single character value
            Must be one of "ot:originallabel", "ot:ottid", or "ot:otttaxonname".
            "ot:originallabel" returns the tree with tip labels as it was originally
              submitted to phylesystem by a curator.
            "ot:ottid" returns a tree with tip labels corresponding to the matching
              ott id.
            "ot:otttaxonname" returns a tree with tip labels corresponding to the
              matching ott taxon name.
        demand_success : boolean
            Whether to return an error or return a somewhat failed output silently.
        """
        if tree_format not in ["newick", "nexson", "nexus", "object"]:
            raise ValueError('"{}" not recognized as a valid tree_format'.format(tree_format))
        if tree_format == 'object':
            ws_rec = self.ws.study(study_id, demand_success=False)

            def efn(rd):
                nexs = rd['data']
                return ws_rec._to_object_converter.tree_from_nexson(nexs,
                                                                    tree_id=tree_id,
                                                                    label_format=label_format)

            ws_rec._tree_from_response_extractor = efn
        else:
            ws_rec = self.ws.tree(study_id, tree_id, tree_format, label_format, demand_success)
            from .ws_wrapper import extract_content_from_raw_text_method_dict
            if tree_format in ('newick', 'nexus'):
                ws_rec._tree_from_response_extractor = extract_content_from_raw_text_method_dict
        return ws_rec

    def get_otus(self, study_id):
        """
        Get OTUs from a study in the Open Tree of Life Phylesystem.

        Parameters
        ----------
        study_id : single character value
            The study id from Open Tree of Life.
        """
        return self.ws.otus(study_id)

    def conflict_info(self, study_id, tree_id, compare_to='synth'):
        """
        Get node status data from any tree in the Open Tree of Life Phylesystem.

        Parameters
        ----------
        study_id : single character value
            The study id from Open Tree of Life.
        tree_id : single character value
            The tree id of a tree within the study id provided.
        compare_to : a single character value
            Usually, you want this to be 'synth', to compare to the synthetic tree.
            Alternatively, you can compare your tree to any other tree in phylesystem.
        """
        return self.ws.conflict(study_id, tree_id, compare_to, demand_success=False)

    def conflict_str(self, tree_str, compare_to='synth'):
        """
        Get node status data from a newick string tree with ott_ids as labels, following the rough
        format:
        "(('_nd1_ott770315','newick_nd2_ott417950')'_nd3_','_nd4_ott158484')'_nd5';".

        Parameters
        ----------
        tree_str: a tree in 'conflict formatted' newick string
        compare_to : a single character value
            Usually, you want this to be 'synth', to compare to the synthetic tree.
            Alternatively, you can compare your tree to any other tree in phylesystem.
        """
        return self.ws.conflict_from_newick(tree_str, compare_to, demand_success=False)

    def studies_properties(self):
        """
        Get properties that can be used to search across studies and trees in phylesystem.
        """
        return self.ws.studies_properties()

    def find_studies(self, value, search_property, exact=False, verbose=False):
        """
        Get study ids that match a certain value of a given search property.

        Parameters
        ----------
        value : single character value
            The study id from Open Tree of Life.
        search_property : single character value
            Any value from studies_properties.
        exact : boolean

        verbose : boolean
        """
        return self.ws.studies_find_studies(value, search_property=search_property,
                                            exact=exact, verbose=verbose)

    def find_trees(self, value, search_property, exact=False, verbose=False):
        """
        Get trees that match a certain value of a given search property.

        Parameters
        ----------
        value : single character value
            The study id from Open Tree of Life.
        search_property : single character value
            Any value from studies_properties.
        exact : boolean

        verbose : boolean

        Example
        -------


        """
        return self.ws.studies_find_trees(value, search_property=search_property,
                                          exact=exact, verbose=verbose)

    def taxon_info(self, ott_id=None, source_id=None, include_lineage=False,
                   include_children=False, include_terminal_descendants=False):
        """
        Get taxonomic information for a given taxon in the Open Tree taxonomy.

        Parameters
        ----------
        ott_id : single character value
            The OTT id of a taxon.
        source_id : maybe single character value

        include_lineage : boolean

        include_children : boolean

        include_terminal_descendant : boolean
        """
        return self.ws.taxonomy_taxon_info(ott_id=ott_id, source_id=source_id,
                                           include_lineage=include_lineage,
                                           include_children=include_children,
                                           include_terminal_descendants=include_terminal_descendants)

    def taxon_mrca(self, ott_ids=None):
        """
        Get the node corresponding to the most recent commom ancestor (mrca) of
          a taxon in the synthetic Open Tree of Life tree.
        Notes from Luna:
            Does it work with just one id?
            Since it is not always a taxon mrca, should it be called get_mrca?

        Parameters
        ----------
        ott_ids : maybe single character value

        """
        while True:
            call_record = self.ws.taxonomy_mrca(ott_ids=ott_ids)
            if call_record:
                return call_record
            else:
                msgtemplate = 'Call to taxonomy/mrca failed with the message "{}"'
                message = call_record.response_dict['message']
                raise OTWebServicesError(msgtemplate.format(message))
            ## cull_unknown_ids_from_args does not work on taxon_mrca call_record
            # self._cull_unknown_ids_from_args(call_record, [], ott_ids) 

    def taxon_subtree(self, ott_id=None, label_format="name_and_id"):
        """Get a subtree of a particular taxon
        """
        return self.ws.taxonomy_subtree(ott_id=ott_id, label_format=label_format)

    def tnrs_contexts(self):
        """Get a list of taxonomic contexts that can be used to constraint a TNRS match.
        """
        return self.ws.tnrs_contexts()

    def tnrs_infer_context(self, names):
        """Infer taxonomic context for names via a TNRS (Taxonomic Name Resolution Service) match.
        """
        return self.ws.tnrs_infer_context(names)

    def tnrs_match(self, names, context_name=None, do_approximate_matching=False,
                   include_suppressed=False):
        """Match taxon names to Open Tree Taxonomy using TNRS (Taxonomic Name Resolution
             Service).
        """
        return self.ws.tnrs_match_names(names, context_name=context_name,
                                        do_approximate_matching=do_approximate_matching,
                                        include_suppressed=include_suppressed)

    def tnrs_autocomplete(self, name, context_name=None, include_suppressed=False):
        """Taxonomic name resolution service autocomplete
        """
        return self.ws.tnrs_autocomplete_name(name, context_name=context_name,
                                              include_suppressed=include_suppressed)

    def synth_node_info(self, node_ids=None, node_id=None, ott_id=None, include_lineage=False):
        """Get information of a node
        """
        if isinstance(node_ids, str):
            node_ids = [node_ids]
        if ott_id:
            if isinstance(ott_id, str):
                ott_id=ott_id.strip('ott')
                ott_id = int(ott_id)
        return self.ws.tree_of_life_node_info(node_ids=node_ids, node_id=node_id, ott_id=ott_id,
                                              include_lineage=include_lineage)

    def synth_subtree(self, node_id=None, ott_id=None,
                      tree_format="newick", label_format="name_and_id",
                      height_limit=None):
        """Get a subtree
        """
        return self.ws.tree_of_life_subtree(node_id=node_id, ott_id=ott_id,
                                            tree_format=tree_format,
                                            label_format=label_format,
                                            height_limit=height_limit)

    def synth_induced_tree(self, node_ids=None,
                           ott_ids=None, label_format="name_and_id",
                           ignore_unknown_ids=False):
        """Get an induced subtree
        """
        while True:
            call_record = self.ws.tree_of_life_induced_subtree(node_ids=node_ids,
                                                               ott_ids=ott_ids,
                                                               label_format=label_format)
            if call_record:
                return call_record
            if not ignore_unknown_ids:
                msgtemplate = 'Call to tree_of_life/induced_subtree failed with the message "{}"'
                message = call_record.response_dict['message']
                if call_record.response_dict['unknown']:
                    unknown_ids = call_record.response_dict['unknown']
                    message = message + "\nFull list of unknown/unrecognized query ids:\n {} \n".format("\n".join(unknown_ids))
                raise OTWebServicesError(msgtemplate.format(message))
            msgtemplate = 'Call to tree_of_life/induced_subtree failed with the message "{}"'
            self._cull_unknown_ids_from_args(call_record, node_ids, ott_ids)

    def synth_mrca(self, node_ids=None, ott_ids=None, ignore_unknown_ids=True):
        """Get the most recent common ancestor of a group of taxa on the synthetic
             Open Tree of Life
        """
        while True:
            assert (ott_ids or node_ids)
            call_record = self.ws.tree_of_life_mrca(node_ids=node_ids,
                                                    ott_ids=ott_ids)
            if call_record:
                return call_record
            if not ignore_unknown_ids:
                msgtemplate = 'Call to tree_of_life/mrca failed with the message "{}"'
                message = call_record.response_dict['message']
                raise OTWebServicesError(msgtemplate.format(message))
            self._cull_unknown_ids_from_args(call_record, node_ids, ott_ids)
            if not ott_ids or node_ids:
                msgtemplate = 'Call to tree_of_life/mrca failed as all ids were pruned'
                raise OTWebServicesError(msgtemplate)

    # noinspection PyMethodMayBeStatic
    def _cull_unknown_ids_from_args(self, call_record, node_ids, ott_ids, output='err'):
        """Cull unknown ids from arguments
        """
        assert ('unknown' in call_record.response_dict), call_record.response_dict
        unknown_ids = call_record.response_dict['unknown']
        sys.stderr.write("\nUnknown/unrecognized query ids (skipped):\n {} \n".format("\n".join(unknown_ids)))
        for u in unknown_ids:
            if node_ids and u in node_ids:
                node_ids.remove(u)
            else:
                assert u.startswith('ott')
                ui = int(u[3:])
                if ott_ids and (ui in ott_ids):
                    ott_ids.remove(ui)
                if ott_ids and (str(ui) in ott_ids):
                    ott_ids.remove(str(ui))

    def get_ottid_from_gbifid(self, gbif_id):
        """Returns an ott id for a gbif id
        ott_id is set to 'None' if the gbif id is not found in the Open Tree Taxanomy
        """
        assert int(gbif_id)
        gbiftax = "gbif:{}".format(int(gbif_id))
        res = self.taxon_info(source_id=gbiftax)
        if res.status_code == 200:
            ott_id = int(res.response_dict['ott_id'])
            return ott_id
        if res.status_code == 400:
            msgtemplate = 'Call to taxon_info failed with the message "{}"'
            message = res.response_dict['message']
            raise OTWebServicesError(msgtemplate.format(message))

    def get_citations(self, studies):
        """Returns study citations from a list of study or tree ids
        """
        cites = []
        for study in studies:
            if '@' in study:
                studyid = study.split('@')[0]
                treeid = study.split('@')[1]
                opentree_url = "https://tree.opentreeoflife.org/curator/study/view/{}?tab=trees&tree={}"
                opentree_url = opentree_url.format(studyid, treeid)
            else:
                studyid = study
                opentree_url = "https://tree.opentreeoflife.org/curator/study/view/{}".format(studyid)
            studyres = self.find_studies(studyid, search_property='ot:studyId', verbose=True)
            new_cite = studyres.response_dict.get('matched_studies', None)
            if new_cite:
                cites.append(opentree_url + '\n' +
                             new_cite[0].get('ot:studyPublicationReference', '')
                             + '\n' +
                             new_cite[0].get('ot:studyPublication', '') + '\n')
        return "\n".join(cites)
        
    def get_ottid_from_name(self, spp_name):
        """Returns an ott id for a string - requires exact match.
        ott_id is set to 'None' if the name is not found in the Open Tree Txanomy
        """
        res = self.tnrs_match([spp_name], do_approximate_matching=False)
        if res.status_code == 200:
            if len(res.response_dict['results'][0]['matches']) > 0:
                tax = res.response_dict['results'][0]['matches'][0].get('taxon')
                ott_id = int(tax.get('ott_id'))
                return ott_id
            else:
                sys.stderr.write("Exact match to name {} not found in taxonomy.\n".format(spp_name))
                sys.stderr.write("""Try using `resp = OT.tnrs_match(["{}"], do_approximate_matching=True)`\n
                                                   resp.response_dict \n
                                        to find fuzzy matches\n""".format(spp_name))
                return None
        msgtemplate = 'Call to tnrs_match failed with the message "{}"'
        message = res.response_dict['message']
        raise OTWebServicesError(msgtemplate.format(message))

    def get_matchdict_from_taxlist(self, list_of_taxa):
        """
        Input: a list of taxon names
        Returns: matches - a dictionary of name:ott_id
        and failed - a set of the names that were not found.
        """
        matches = dict()
        failed = set()
        for tax in list_of_taxa:
            tax = tax.strip()
            if tax != '':
                ott_id = self.get_ottid_from_name(tax)
                if ott_id is None:
                    failed.add(tax)
                else:
                    matches[tax] = 'ott{}'.format(ott_id)
        return matches, failed


#!/usr/bin/env python3

class SourceSupportCollection(object):
    def __init__(self, supp_descrip, src_id_map):
        pass

class DictWrapper(object):
    def __init__(self, raw_dict):
        self._raw_dict = raw_dict

    def _move_to_attribute(self, attr_name, conv_fn=None):
        val = self._raw_dict.get(attr_name)
        if (val is not None) and (conv_fn is not None):
            val = conv_fn(val)
        setattr(self, '_{}'.format(attr_name), val)
        return val

    def _access_or_move(self, attr_name, conv_fn=None):
        wattr = '_{}'.format(attr_name)
        val = getattr(self, wattr)
        if val is None:
            return self._move_to_attribute(attr_name, conv_fn=conv_fn)
        return val

def wrap_list_elements(container, wrapper_class):
    return [wrapper_class(i) for i in container]

def wrap_ott_id_list_as_taxon(container):
    return wrap_list_elements([{'ott_id': i} for i in container], OTTaxonRef)

def wrap_list_of_dicts_as_ot_taxon_ref(container):
    return wrap_list_elements(container, OTTaxonRef)

class OTTaxonRef(DictWrapper):
    def __init__(self, taxon_dict):
        super(OTTaxonRef, self).__init__(taxon_dict)
        self._name = None
        self._ott_id = None
        self._rank = None
        self._tax_sources = None
        self._unique_name = None
        self._flags = None
        self._is_suppressed = None
        self._is_suppressed_from_synth = None
        self._source = None
        self._terminal_descendants = None
        self._lineage = None
        self._children = None
        self._fetched = set()

    def _access_fetch_or_move(self, attr_name, conv_fn=None, **kwargs):
        v = self._access_or_move(attr_name, conv_fn=conv_fn)
        if v is None and (attr_name not in self._fetched):
            from .ot_object import default_open_tree_obj
            OT = default_open_tree_obj()
            rd = OT.taxon_info(self.ott_id, **kwargs)
            self._raw_dict.update(rd.response_dict)
            v = self._access_or_move(attr_name, conv_fn=conv_fn)
            self._fetched.add(attr_name)
        return v

    @property
    def name(self):
        return self._access_or_move('name')

    @property
    def ott_id(self):
        return self._access_or_move('ott_id')

    @property
    def rank(self):
        return self._access_or_move('rank')

    @property
    def tax_sources(self):
        return self._access_or_move('tax_sources')

    @property
    def unique_name(self):
        return self._access_or_move('unique_name')

    @property
    def flags(self):
        return self._access_fetch_or_move('flags')

    @property
    def is_suppressed(self):
        return self._access_fetch_or_move('is_suppressed')

    @property
    def is_suppressed_from_synth(self):
        return self._access_fetch_or_move('is_suppressed_from_synth')

    @property
    def source(self):
        return self._access_fetch_or_move('source')

    @property
    def lineage(self):
        return self._access_fetch_or_move('lineage',
                                          include_lineage=True,
                                          conv_fn=wrap_list_of_dicts_as_ot_taxon_ref)

    @property
    def children(self):
        return self._access_fetch_or_move('children',
                                          include_children=True,
                                          conv_fn=wrap_list_of_dicts_as_ot_taxon_ref)

    @property
    def terminal_descendants(self):
        return self._access_fetch_or_move('terminal_descendants',
                                          include_terminal_descendants=True,
                                          conv_fn=wrap_ott_id_list_as_taxon)

class SynthNodeReference(object):
    def __init__(self, node_descrip_dict):
        ndd = node_descrip_dict
        src_id_map = ndd.get('source_id_map', {})
        self.node_id = ndd['node_id']
        self.num_tips = ndd['num_tips']
        self.query = ndd.get('query')
        td = ndd.get('taxon', None)
        self.taxon = None if td is None else OTTaxonRef(td)
        SSC = SourceSupportCollection
        self.supported_by = SSC(ndd.get('supported_by', {}), src_id_map)
        self.conflicts_with = SSC(ndd.get('conflicts_with', {}), src_id_map)
        self.supported_by = SSC(ndd.get('supported_by', {}), src_id_map)
        self.resolves = SSC(ndd.get('resolves', {}), src_id_map)
        self.partial_path_of = SSC(ndd.get('partial_path_of', {}), src_id_map)
        self.terminal = SSC(ndd.get('terminal', {}), src_id_map)
        self._lineage = ndd.get('lineage', [])
        self._lineage_converted = False

    @property
    def lineage(self):
        if self._lineage and (not self._lineage_converted):
            self._lineage = [SynthNodeReference(i) for i in self._lineage]
            self._lineage_converted = True
        return self._lineage



#!/usr/bin/env python3
"""Miscellaneous light-weight functions for common operations when working with Open Tree data"""
import logging
import sys

_TAXONOMY_WIKI_URL = "https://github.com/OpenTreeOfLife/reference-taxonomy/wiki"
_TAXON_FLAGS_WIKI_URL = "{}/Taxon-flags".format(_TAXONOMY_WIKI_URL)
_SUPPRESSED_TAXON_FLAGS_WIKI_URL = _TAXON_FLAGS_WIKI_URL + '#flags-leading-to-taxa-being-unavailable-for-tnrs'


def get_suppressed_taxon_flag_expl_url():
    """Returns the current URL describing taxon flags that lead to suppression"""
    return _SUPPRESSED_TAXON_FLAGS_WIKI_URL


def ott_str_as_int(o):
    """Returns the OTT Id `o` as an integer if `o` is an integer or a string starting with ott (case-insensitive).

    Raises a ValueError if the string does not match ^(OTT)?[0-9]+$
    """
    if isinstance(o, int):
        return o
    if o.startswith('ott'):
        return int(o[3:])
    try:
        return int(o)
    except:
        if o.lower().startswith('ott'):
            return int(o[3:])
        raise


def _create_link_from_node_info_conf_key_value_pair(key, value):
    """Takes the key value pairs contained in the ToL/node_info call and returns a URL for the node"""
    if key.lower().startswith('ott'):
        value = ott_str_as_int(value)
        return "https://tree.opentreeoflife.org/taxonomy/browse?id={}".format(value)
    study_id, tree_id = key.split('@')
    node_id = str(value)
    if node_id.startswith('ott'): ##todo Check if link is possible
        logging.debug("node highlighting will not work for ({}, {}, {})\n".format(study_id, tree_id, node_id))
        # # This node does not occur in the input tree, it is a result of exemplification...
        # #   link to its parent node in the input study
        # inp_tree_nexson = OT.get_tree(study_id=study_id, tree_id=tree_id)
        # tax_lineage = OT.taxon_info(ott_id=ott_str_as_int(node_id), include_lineage=True)
        # anc_ids = [i.ott_id for i in tax_lineage.tree.postorder_node_iter()]
        # matches = []
        # for leaf in inp_tree_nexson.tree.leaf_node_iter():
        #     if leaf.otu.ott_id and leaf.otu.ott_id in anc_ids:
        #         matches.append((anc_ids.index, id(leaf), leaf))
        # matches.sort()
        # if not matches:
        #     logging.warning("Could not find input node for ({}, {}, {})\n".format(study_id, tree_id, node_id))
        # else:
        #     node_id = matches[0][-1].node_id
        #
    tmp = "https://tree.opentreeoflife.org/curator/study/view/{s}?tab=trees&tree={t}&node={n}"
    return tmp.format(s=study_id, t=tree_id, n=node_id)


def _format_link_to_input_tree(link_dict):
    """Returns a list of lines with node links. `link_dict` that is from a ToL/node_info response's support fields"""
    lines = []
    keys = list(link_dict.keys())
    keys.sort()
    for key in keys:
        url = _create_link_from_node_info_conf_key_value_pair(key, link_dict[key])
        if key.lower().startswith('ott'):
            line = "The Open Tree taxonomy ({}): {}".format(key, url)
        else:
            line = "input phylogeny at {}".format(url)
        lines.append(line)
    return lines


def write_node_info_links_to_input_trees(blob, out=sys.stdout):
    """Writes a summary of the support/conflict info from a ToL/node_info call response `blob` to stream `out`"""
    x = blob.get('supported_by')
    if x:
        out.write("The existence of this node in the synthetic tree is supported by:\n")
        for n, line in enumerate(_format_link_to_input_tree(x)):
            out.write(" {c:3d}: {l}\n".format(c=n + 1, l=line))
    x = blob.get('partial_path_of')
    if x:
        out.write("The branch to this node in the synthetic tree is part of a path that display the node(s):\n")
        for n, line in enumerate(_format_link_to_input_tree(x)):
            out.write(" {c:3d}: {l}\n".format(c=n + 1, l=line))
    x = blob.get('resolves')
    if x:
        out.write("The branch to this node in the synthetic tree would resolve the polytomy at node(s):\n")
        for n, line in enumerate(_format_link_to_input_tree(x)):
            out.write(" {c:3d}: {l}\n".format(c=n + 1, l=line))
    x = blob.get('terminal')
    if x:
        out.write("The branch to this node in the synthetic tree is part of a terminal edge to node(s):\n")
        for n, line in enumerate(_format_link_to_input_tree(x)):
            out.write(" {c:3d}: {l}\n".format(c=n + 1, l=line))
    x = blob.get('conflicts_with')
    if x:
        out.write("The existence of this node in the synthetic tree is conflicts with:\n")
        for n, line in enumerate(_format_link_to_input_tree(x)):
            out.write(" {c:3d}: {l}\n".format(c=n + 1, l=line))




import unittest

from opentree import OT

bad_study_id = "pg_873"
study_id = "ot_350"
synth_spp = ['770769','638559','336228','638539','283034']
expected_tips = ['Sterculia tragacantha ott770769', 'Hildegardia barteri ott336228', 'Sterculia balanghas ott638559', 'Firmiana simplex ott283034', 'Firmiana malayana ott638539']
tax_mrca = 996482
spp_name = "Bos grunniens"
bg_tax = 381164
conf_newick_str = "(('_nd1_ott770315','_nd2_ott417950')'_nd3_','_nd4_ott158484')'_nd5';"

class TestPhyscraperDeps(unittest.TestCase):
    def test_demands_id_arg(self):
        with self.assertRaises(TypeError):
            OT.get_study()

    def test_get_study_fail(self):
        study = OT.get_study(bad_study_id)
        assert study.response_dict == {'description': 'Study #{} GET failure'.format(bad_study_id), 'error': 1}

    def test_get_study(self):
        study = OT.get_study(study_id)
        assert 'data' in study.response_dict

    def test_synth_induced(self):
        tre = OT.synth_induced_tree(ott_ids=synth_spp).tree
        leaves = [leaf.taxon.label for leaf in OT.synth_induced_tree(ott_ids=synth_spp).tree.leaf_nodes()]
        assert leaves == expected_tips

    def test_taxon_subtree(self):
        resp = OT.taxon_subtree(tax_mrca)
        taxleaves = [leaf.taxon.label for leaf in resp.tree.leaf_nodes()] 
        assert len(taxleaves) == 585


    def test_taxon_mrca(self):
        mrca = OT.taxon_mrca(synth_spp).response_dict['mrca']['ott_id']
        assert mrca == tax_mrca

    def test_synth_mrca(self):
        mrca_node = OT.synth_mrca(ott_ids = synth_spp).response_dict
        assert u'taxon' in mrca_node['mrca'].keys()
        tax_id = mrca_node['mrca'][u'taxon'][u'ott_id']
        assert tax_id ==  279960

    def test_get_taxon_info(self):
        call = OT.tnrs_match([spp_name], do_approximate_matching=True)
        res = call.response_dict['results'][0]
        assert res['matches'][0]['matched_name'] == 'Bos grunniens'
        assert res['matches'][0]['taxon']['ott_id'] ==  bg_tax

    def test_ottid_from_name(self):
        ottid = OT.get_ottid_from_name('Bos grunniens')
        assert ottid == bg_tax

    def test_conflict_str(self):
        resp = OT.conflict_str(conf_newick_str, 'ott').response_dict
        assert resp['nd3']['status'] == 'conflicts_with'
    
    def test_find_trees(self):
        phylesystem_studies_resp = OT.find_trees(bg_tax, search_property ='ot:ottId')
        matches = [study['ot:studyId'] for study in phylesystem_studies_resp.response_dict['matched_studies']]
        assert 'ot_409' in matches

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import dendropy
import re


def get_object_converter(object_conversion_schema):
    if object_conversion_schema.lower() == 'dendropy':
        return DendropyConvert()
    raise ValueError('Currently only conversion to DendroPy objects is supported.')


_name_gap_ott_num = re.compile(r'^(.+)[ _]ott(\d+)$')


def _decorate_taxa_in_taxon_namespace_by_parsing_labels(tree):
    taxon_namespace = tree.taxon_namespace
    for nd in tree.preorder_node_iter():
        if (nd.taxon is not None) or (not nd.label):
            continue
        m = _name_gap_ott_num.match(nd.label)
        if m:
            nd.taxon = taxon_namespace.new_taxon(nd.label)

    for taxon in taxon_namespace:
        label = taxon.label
        m = _name_gap_ott_num.match(label)
        if m:
            name = m.group(1).strip()
            if name and not hasattr(taxon, 'ott_taxon_name'):
                taxon.ott_taxon_name = name
            if not hasattr(taxon, 'ott_id'):
                taxon.ott_id = int(m.group(2))


# noinspection PyMethodMayBeStatic
class DendropyConvert(object):
    """
    Class to convert newicks to dendropy objects
    """

    def tree_from_newick(self, newick, suppress_internal_node_taxa=False, **kwargs):
        tree = dendropy.Tree.get(data=newick, schema="newick",
                                 suppress_internal_node_taxa=suppress_internal_node_taxa, **kwargs)
        _decorate_taxa_in_taxon_namespace_by_parsing_labels(tree)
        return tree

    def tree_list_from_newicks(self, newick_list, suppress_internal_node_taxa=False, **kwargs):
        concat = '\n'.join(newick_list)
        tree_list = dendropy.TreeList.get(data=concat,
                                          schema="newick",
                                          suppress_internal_node_taxa=suppress_internal_node_taxa,
                                          **kwargs)
        for tree in tree_list:
            _decorate_taxa_in_taxon_namespace_by_parsing_labels(tree)
        return tree_list

    def taxon_namespace_and_id_dict_from_nexson_otus_obj(self, otus_obj, otus_id):
        tn = dendropy.TaxonNamespace(label=otus_id)
        id_to_taxon = {}
        for oid, otu_obj in otus_obj.items():
            dt = tn.new_taxon(oid)
            if oid in id_to_taxon:
                raise ValueError('otu id "{}" repeated'.format(oid))
            id_to_taxon[oid] = dt
            dt.otu = oid
            dt.ott_taxon_name, dt.ott_id, dt.original_label = None, None, None
            for meta_key, meta_v in otu_obj.items():
                if meta_key == '^ot:ottTaxonName':
                    dt.ott_taxon_name = meta_v
                elif meta_key == '^ot:originalLabel':
                    dt.original_label = meta_v
                elif meta_key == '^ot:ottId':
                    dt.ott_id = meta_v
        return tn, id_to_taxon

    def tree_from_nexson(self, nexson, tree_id, label_format="ot:originallabel"):
        to_taxon_attr = {
            "ot:originallabel": "original_label",
            "ot:ottid": "ott_id",
            "ot:otttaxonname": "ott_taxon_name",
            "id": "ott_id",
            "name": "ott_taxon_name"
        }
        taxon_attr = to_taxon_attr[label_format.lower()]
        nexml = nexson['nexml']
        trees_sets_by_id = nexml.get('treesById', {})
        otu_set_id, tree_obj = None, None
        for tree_set_id, tree_set in trees_sets_by_id.items():
            otu_set_id = tree_set.get('@otus')
            tbi = tree_set.get('treeById', {})
            tree_obj = tbi.get(tree_id)
            if tree_obj:
                break
        if tree_obj is None:
            raise KeyError('Tree with id "{}" not found in NexSON'.format(tree_id))
        if otu_set_id is None:
            raise KeyError('Tree set missing "@otus" property')
        otus_by_id = nexml['otusById']
        otu_set = otus_by_id[otu_set_id]
        #        if len(otu_set) != 1:
        #            raise ValueError('expecting just  "otuById" in OTUs object')
        #       # Hmmm. Some have otu_set.keys() = dict_keys(['@label', '^skos:historyNote', 'otuById']). Seems fine.
        obi = otu_set["otuById"]
        tn, id2taxon = self.taxon_namespace_and_id_dict_from_nexson_otus_obj(obi, otu_set_id)
        for taxon in tn:
            tl = getattr(taxon, taxon_attr, None)
            if tl is not None:
                taxon.label = tl
        tree = dendropy.Tree(label=tree_obj.get('@label', tree_id), taxon_namespace=tn)
        tt = tree_obj["@xsi:type"]
        if tt == 'nex:FloatTree':
            tree.length_type = float
        root_node_id = tree_obj["^ot:rootNodeId"]
        spec_root_id = tree_obj["^ot:specifiedRoot"]
        if spec_root_id:
            assert spec_root_id == root_node_id
        non_annotations = frozenset(["^ot:inGroupClade", "^ot:rootNodeId", "^ot:specifiedRoot"])
        for key, value in tree_obj.items():
            if key in non_annotations:
                continue
            if key.startswith('^ot:'):
                tree.annotations.add_new(name=key[4:], value=value)
        nex_nds = tree_obj['nodeById']
        edges_by_src = tree_obj['edgeBySourceId']
        nd_id2dend = {}
        nexson_nd = nex_nds[root_node_id]
        _proc_nd(tree.seed_node, nexson_nd, id2taxon, root_node_id, nd_id2dend)
        to_proc = [tree.seed_node]
        while to_proc:
            next_nd = to_proc.pop(0)
            edges_dict = edges_by_src.get(next_nd.id, {})
            for edge_id, e_dict in edges_dict.items():
                new_nd = tree.node_factory()
                assert e_dict['@source'] == next_nd.id
                dest_id = e_dict['@target']
                _proc_nd(new_nd, nex_nds[dest_id], id2taxon, dest_id, nd_id2dend)
                next_nd.add_child(new_nd)
                edge_len = e_dict.get('@length')
                new_nd.edge.id = edge_id
                if edge_len is not None:
                    new_nd.edge.length = edge_len
                to_proc.append(new_nd)
        return tree


def _proc_nd(node, nexson_nd, id2taxon, node_id, nd_id2dend):
    node.id = node_id
    nd_id2dend[node_id] = node
    _decorate_nd(node, nexson_nd, id2taxon)


def _decorate_nd(node, nexson_nd, id2taxon):
    otu_id = nexson_nd.get('@otu')
    if otu_id:
        node.taxon = id2taxon[otu_id]
    node.label = node.id



import unittest

from opentree import OT
from opentree.ws_wrapper import OTWebServicesError

study_id = 'ot_1979'
tree_id = 'tree1'
bos = 1066581
homo = 770309

class TestOT(unittest.TestCase):
    def test_synth_files(self):
        FS = OT.files_server
        scaff_tree = OT.get_subproblem_scaffold_tree('opentree12.3')
        sub_size = OT.get_subproblem_size_info('opentree12.3')


    def test_subproblems_fail(self):
        with self.assertRaises(OTWebServicesError):
            ## Bos is not in the synth tree
            sol = OT.get_subproblem_solution('opentree12.3', bos)
            sol_trees = OT.get_subproblem_trees('opentree12.3', bos)
            rev_sol = OT.get_reversed_subproblem_solution('opentree12.3', bos)

    def test_subproblems(self):
        sol = OT.get_subproblem_solution('opentree12.3', homo)
        sol_trees = OT.get_subproblem_trees('opentree12.3', homo)
        rev_sol = OT.get_reversed_subproblem_solution('opentree12.3', homo)

    def test_about(self):
        ret = OT.about()
        assert 'taxonomy_about', 'synth_tree_about' in ret

    def test_get_tree_newick(self):
        res = OT.get_tree(study_id=study_id, tree_id=tree_id, tree_format='newick')
        nwk = res.response_dict['content'].decode("utf-8")
        assert isinstance(nwk, str)

    def test_get_tree_bad_format(self):
        with self.assertRaises(ValueError):
            res = OT.get_tree(study_id=study_id, tree_id=tree_id, tree_format='newrk')

    def test_get_tree(self):
        res =  OT.get_tree(study_id=study_id, tree_id=tree_id, label_format='ot:otttaxonname',tree_format='nexus') 
        nex = res.response_dict['content'].decode("utf-8")
        assert nex.startswith("#NEXUS")

    def test_get_tree_object(self):
        res =  OT.get_tree(study_id=study_id, tree_id=tree_id, label_format='ot:ottid',tree_format='object') 
        tree_dict = res.response_dict['data']
        assert isinstance(tree_dict, dict)

    def test_get_otus(self):
        res =  OT.get_otus(study_id=study_id) 
        otu_dict = res.response_dict['otus1']['otuById']
        assert len(otu_dict) == 45

    def test_conflict(self):
        res = OT.conflict_info(study_id=study_id, tree_id=tree_id, compare_to='synth')
        assert 'node5' in res.response_dict

    def test_properties(self):
        properties = OT.studies_properties().response_dict
        assert 'ot:curatorName', 'ot:curatorName' in properties['study_properties']
        assert 'ot:ottid', 'ot:curatorName' in properties['tree_properties']

    def test_find_studies(self):
       res = OT.find_studies("Ilex", "ot:focalCladeOTTTaxonName")
       assert  "ot_1984" in [match['ot:studyId'] for match in res.response_dict["matched_studies"]]

    def test_taxon_subtree(self):
        res = OT.taxon_subtree(ott_id=bos)
        nwk = res.response_dict['newick']
        assert isinstance(nwk, str)
        curl = res.curl_call
        tax = res.taxon

    def test_taxon_info(self):
        res = OT.taxon_info(ott_id = bos).response_dict
        assert 'Taurus' in res['synonyms']

    def test_taxon_mrca(self):
        res = OT.taxon_mrca(ott_ids = [bos, homo]).response_dict
        assert res['mrca']['name'] == 'Boreoeutheria'

    def test_taxon_mrca_fail(self):
        with self.assertRaises(OTWebServicesError):
            res = OT.taxon_mrca(ott_ids = [bos, homo, 9999999])

    def test_contexts(self):
        contexts = OT.tnrs_contexts()
        amph = OT.tnrs_infer_context(["Bufo", "Rana", "Hyla"])
        contxt  = amph.response_dict['context_name']
        assert contxt in contexts.response_dict['ANIMALS']

    def test_tnrs_match(self):
        matches = OT.tnrs_match(["Bufo", "Rana", "Hyla"])

    def test_gbif_to_ott(self):
        matchid = OT.get_ottid_from_gbifid(2441017)
        assert matchid == bos
        with self.assertRaises(OTWebServicesError):
            matchid = OT.get_ottid_from_gbifid(999999999)

    def test_get_citations(self):
        cites = OT.get_citations(studies = ['ot_1000@tree1', 'ot_1984'])

    def test_matchdict(self):
        matches, failed = OT.get_matchdict_from_taxlist(['Homo', 'Bos', 'Meep'])
        assert matches['Bos'] == 'ott{}'.format(bos)
        assert 'Meep' in failed


if __name__ == '__main__':
    unittest.main()



import unittest

from opentree import taxonomy_helpers, util, OT
from dendropy import Tree

ott_loc = "./"
corr_tax_path = ott_loc + "/ott3.2"

class TestTaxonomyHelpers(unittest.TestCase):
    def test_get_taxonomy(self):
        tax_path = taxonomy_helpers.download_taxonomy_file(loc = ott_loc, version = "3.2")
        assert tax_path == ott_loc + "/ott3.2", tax_path
    def test_get_forward(self):
        tax_path = taxonomy_helpers.download_taxonomy_file(loc = ott_loc, version = "3.2")
        ott_ids = ['ott5859962', 'ott773110']
        forwards_file = tax_path + "/forwards.tsv"
        taxonomy_helpers.get_forwards_dict(forwards_file=forwards_file)
    def test_clean_taxonomy(self):
        cfi = taxonomy_helpers.clean_taxonomy_file(taxonomy_file="{}/taxonomy.tsv".format(corr_tax_path))
    def test_get_by_rank(self):
        ids = taxonomy_helpers.get_ott_ids_for_rank(rank="family", taxonomy_file="{}/taxonomy.tsv".format(corr_tax_path), synth_only = False)
        assert len(ids) == 21979, len(ids)
    def test_get_by_group(self):
        aves = taxonomy_helpers.get_ott_ids_for_group(group_ott_id=81461)
        assert len(aves) == 27465, len(aves)
        aves_synth = taxonomy_helpers.get_ott_ids_for_group(group_ott_id=81461, synth_only = True)
        assert len(aves_synth) == 23618, len(aves_synth)
    def test_rank_in_taxon(self):
        bird_families = taxonomy_helpers.get_ott_ids_group_and_rank(group_ott_id=81461, rank='family', synth_only = False, taxonomy_file="{}/taxonomy.tsv".format(corr_tax_path))
        assert len(bird_families) == 390, len(bird_families)
        bird_families_synth = taxonomy_helpers.get_ott_ids_group_and_rank(group_ott_id=81461, rank='family', taxonomy_file="{}/taxonomy.tsv".format(corr_tax_path))
        assert len(bird_families_synth) == 196, len(bird_families_synth)
    def test_relabel(self): ## JUST PUT A LIST OF IDS TO SIMPLIFY
        jetz = OT.get_tree(study_id='ot_809', tree_id = 'tree1', tree_format="newick", label_format="ot:ottId")
        jetz_tree = Tree.get(string=jetz.response_dict['content'].decode(), schema='newick', suppress_internal_node_taxa=True, suppress_leaf_node_taxa=True)
        tips = [tip.label for tip in jetz_tree.leaf_node_iter()]
        ott_ids =  set()
        for tip in tips:
            try:
                ott_ids.add(int(tip))
            except:
                pass
        ret = taxonomy_helpers.labelled_induced_synth(ott_ids = list(ott_ids), label_format='name')
        tips = [tip.taxon.label for tip in ret['labelled_tree'].leaf_node_iter() if tip.taxon]
        assert len(tips) == 6624
        
        ret = taxonomy_helpers.labelled_induced_synth(ott_ids = list(ott_ids), label_format='name_and_id')
        nodes = [node.taxon.label for node in ret['labelled_tree'] if node.taxon]
        assert 'MRCA of taxa in Amazona auropalliata_ott1118 Amazona oratrix_ott1119' in nodes, nodes



import unittest
import subprocess
from opentree.ot_command_line_tool import (OTCommandLineTool,
                                          process_ott_and_node_id_list_args,
                                          process_ott_or_node_id_arg)
import sys

class TestCLI(unittest.TestCase):
    def test_diagnose_solution(self):
        url = 'https://tree.opentreeoflife.org/curator/study/view/ot_1344?tab=trees&tree=Tr105486&node=Tn16531763'
        p = subprocess.run(["./examples/diagnose_solution_for.py",
                            "--ott-id", "1066581"],  
                            stdout=subprocess.PIPE, 
                            input='\n'.encode()) #input keeps it from hanging waiting for a subproblem id

        assert url in p.stdout.decode('utf8'), sys.executable       

    def test_cli_obj(self):
        sys.argv = ["/examples/diagnose_solution_for.py", "--ott-id", "1066581"]
        cli = OTCommandLineTool(usage="test_example",
                                common_args=("ott-id",))
        cli.parse_cli()

    def test_arg_parse_both(self):
        sys.argv = ["/examples/synth_induced_subtree.py", "--ott-ids", "770309,913932", "--node-ids", "mrcaott354607ott374748"]
        cli = OTCommandLineTool(usage='Display taxonomy and synthetic tree information '
                                  'returned by the "about" API calls.',
                            common_args=("ott-ids", "node-ids"))
        OT, args = cli.parse_cli()
        ott_id_list, node_id_list = process_ott_and_node_id_list_args(args)

    def test_arg_parse_or(self):
        sys.argv = ["/examples/synth_induced_subtree.py", "--ott-id", "913932",]
        cli = OTCommandLineTool(usage='Display taxonomy and synthetic tree information '
                                  'returned by the "about" API calls.',
                            common_args=("ott-id", "node-id"))
        OT, args = cli.parse_cli()
        process_ott_or_node_id_arg(args)



#! /usr/bin/env python3

##############################################################################
#  Adapted from: DendroPy Phylogenetic Computing Library (2010 Jeet Sukumaran
#  and Mark T. Holder)
##############################################################################

"""
opentree testing suite.
"""

import unittest
import re
import os
import logging


def get_test_file_names():
    """Get list of test file names."""
    path = os.path.dirname(__file__)
    default_tests = _get_test_file_names_from_dir(path, 'opentree.test.')
    TEST_PYTHON_OPENTREE_DIR = os.environ.get('TEST_PYTHON_OPENTREE_DIR')
    if TEST_PYTHON_OPENTREE_DIR and os.path.exists(TEST_PYTHON_OPENTREE_DIR):
        atd = os.path.join(TEST_PYTHON_OPENTREE_DIR, 'testopentree')
        if not os.path.isdir(atd):
            logging.warning('TEST_PYTHON_OPENTREE_DIR/testopentree at "{}" does not exist'.format(atd))
        else:
            default_tests.extend(_get_test_file_names_from_dir(atd, 'testopentree.'))
    return default_tests

def _get_test_file_names_from_dir(dirpath, pref):
    files = os.listdir(dirpath)
    t = []
    pat = re.compile(r'^test.*\.py$')
    for f in files:
        if pat.match(f):
            rp = pref + f[:-3]  # [:-3] to strip ".py"
            t.append(rp)
    return t


def get_test_suite(test_file_names=None):
    """
    Creates a unittest.TestSuite from all of the modules in
    `opentree.test`. Right now, assumes (a) no subdirectories (though
    this can easily be accommodated) and (b) every test to be run is
    sitting in a module with a file name of 'test*.py', and, conversely,
    every file with a name of 'test*.py' has test(s) to be run.
    """
    if test_file_names is None:
        test_file_names = get_test_file_names()
    tests = unittest.defaultTestLoader.loadTestsFromNames(test_file_names)
    return unittest.TestSuite(tests)


def run():
    """Runs all of the unittests"""
    runner = unittest.TextTestRunner()
    runner.run(get_test_suite())


if __name__ == "__main__":
    run()



import unittest

from opentree import OT, util

class TestUtil(unittest.TestCase):
    def test_taxon_flag(self):
        corr_url = 'https://github.com/OpenTreeOfLife/reference-taxonomy/wiki/Taxon-flags#flags-leading-to-taxa-being-unavailable-for-tnrs'

        util.get_suppressed_taxon_flag_expl_url()



        assert util.get_suppressed_taxon_flag_expl_url() == corr_url, util.get_suppressed_taxon_flag_expl_url()


        assert util.ott_str_as_int('23') == 23

        corr_ott_link = 'https://tree.opentreeoflife.org/taxonomy/browse?id=123'
        assert util._create_link_from_node_info_conf_key_value_pair('ott', 123) == corr_ott_link

        corr_study_link = 'https://tree.opentreeoflife.org/curator/study/view/ot_1979?tab=trees&tree=tree1&node=node5'
        assert util._create_link_from_node_info_conf_key_value_pair('ot_1979@tree1', 'node5') == corr_study_link

        blob = OT.synth_node_info(node_id='mrcaott354607ott374748').response_dict
        util.write_node_info_links_to_input_trees(blob)


        blob = OT.synth_node_info(node_id='mrcaott177ott29310').response_dict
        util.write_node_info_links_to_input_trees(blob)
        

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
__version__ = "0.0.4"  # sync with setup.py

from .object_conversion import get_object_converter
from .ws_wrapper import (OTWebServicesError,
                         OTClientError,
                         WebServiceRunMode,
                         WebServiceWrapper,
                         )
from .ot_ws_wrapper import OTWebServiceWrapper
from .ot_command_line_tool import (OTCommandLineTool,
                                   process_ott_and_node_id_list_args,
                                   process_ott_or_node_id_arg)
from .ot_object import OpenTree
from .util import (get_suppressed_taxon_flag_expl_url,
                   ott_str_as_int,
                   write_node_info_links_to_input_trees)

# Default-configured wrapper
OT = OpenTree()



#!/usr/bin/env python3
# Class for direct mapping between Open Tree API -> python object

from .object_conversion import get_object_converter
from .ws_wrapper import (WebServiceRunMode,
                         WebServiceWrapper,
                         )
from datetime import datetime
import urllib.parse


# noinspection PyMethodMayBeStatic
class OTWebServiceWrapper(WebServiceWrapper):
    """This class provides a wrapper to the Open Tree of Life web service methods.
    Actual HTTP calls are handled by methods implemented in the base class for clarity of this code.
    API method calls will be mappable to methods in this class. The methods implemented here do argument checking
    and conversion of the returned JSON to more usable objects.
    """

    def __init__(self, api_endpoint, run_mode=WebServiceRunMode.RUN):
        WebServiceWrapper.__init__(self, api_endpoint, run_mode=run_mode)
        self.to_object_converter = get_object_converter('dendropy')

    def _one_and_only_one(self, api_method, param_dict):
        num = sum([1 if bool(v) else 0 for v in param_dict.values()])
        if num == 1:
            for k, v in param_dict.items():
                if v:
                    return k, v
        sl = list(param_dict.keys())
        sl.sort()
        c = '", "'.join(sl)
        raise ValueError('Exactly 1 of "{}" must be provided for a call to {}'.format(c, api_method))

    def studies_properties(self):
        return self._call_api('studies/properties')

    def study(self, study_id, demand_success=False):
        url = 'study/{}'.format(urllib.parse.quote(study_id))
        data = {'output_nexml2json': '1.2.1'}
        return self._call_api(url, http_method='GET', data=data, demand_success=demand_success)

    def tree(self, study_id, tree_id, tree_format="", label_format="ot:originallabel", demand_success=False):
        ext_dict = {'nexus': '.nex', 'newick': '.tre', 'nexson': ''}
        url = 'study/{}/tree/{}{}/?tip_label={}'.format(urllib.parse.quote(study_id), urllib.parse.quote(tree_id),
                                                        ext_dict[tree_format], label_format)
        return self._call_api(url, http_method='GET', demand_success=demand_success)

    def otus(self, study_id, demand_success=False):
        url = 'study/{}/otus'.format(urllib.parse.quote(study_id))
        return self._call_api(url, http_method='GET', demand_success=demand_success)

    # TODO for Luna :) done!
    def conflict(self, study_id, tree_id, compare_to="synth", demand_success=False):
        url = 'conflict/conflict-status?tree1={}%23{}&tree2={}'.format(urllib.parse.quote(study_id),
                                                                       urllib.parse.quote(tree_id), compare_to)
        return self._call_api(url, http_method='GET', demand_success=demand_success)

    def conflict_from_newick(self, input_newick, compare_to, demand_success=False):
        assert(compare_to in ['synth', 'ott'])
        d = {"tree1newick": input_newick,
             "tree2": compare_to }
        return self._call_api('conflict/conflict-status', data=d, demand_success=demand_success)

    def studies_find_studies(self, value, search_property, exact=False, verbose=False):
        d = {"property": str(search_property), "value": str(value), "exact": bool(exact), "verbose": bool(verbose)}
        return self._call_api('studies/find_studies', data=d, demand_success=False)

    def studies_find_trees(self, value, search_property, exact=False, verbose=False):
        d = {"property": str(search_property), "value": str(value), "exact": bool(exact), "verbose": bool(verbose)}
        return self._call_api('studies/find_trees', data=d, demand_success=False)

    def taxonomy_about(self):
        return self._call_api('taxonomy/about')

    def taxonomy_mrca(self, ott_ids):
        if not ott_ids:
            raise ValueError("ott_ids must be provided to a taxonomy/mrca call")
        d = {"ott_ids": [int(i) for i in ott_ids]}
        return self._call_api('taxonomy/mrca', data=d, demand_success=False)

    def taxonomy_taxon_info(self, ott_id=None, source_id=None, include_lineage=False,
                            include_children=False, include_terminal_descendants=False):
        if isinstance(ott_id, str):
            ott_id=ott_id.strip('ott')
            ott_id = int(ott_id)
        cdict = {"source_id": source_id, "ott_id": ott_id}
        id_spec, id_arg = self._one_and_only_one("taxonomy/taxon_info", cdict)
        d = {id_spec: id_arg,
             "include_lineage": include_lineage,
             "include_children": include_children,
             "include_terminal_descendants": include_terminal_descendants, }
        return self._call_api('taxonomy/taxon_info', data=d, demand_success=False)

    def taxonomy_subtree(self, ott_id, label_format=None):
        d = {"ott_id": int(ott_id)}
        if label_format:
            d["label_format"] = str(label_format)
        return self._call_api('taxonomy/subtree', data=d, demand_success=False)

    def tnrs_autocomplete_name(self, name, context_name=None, include_suppressed=False):
        d = {"name": str(name), "include_suppressed": bool(include_suppressed), }
        if context_name is not None:
            d['context_name'] = str(context_name)
        return self._call_api('tnrs/autocomplete_name', data=d, demand_success=False)

    def tnrs_contexts(self):
        return self._call_api('tnrs/contexts', data=None, demand_success=False)

    def tnrs_infer_context(self, names):
        d = {"names": [str(i) for i in names]}
        return self._call_api('tnrs/infer_context', data=d, demand_success=False)

    def tnrs_match_names(self, names, context_name=None, do_approximate_matching=False, include_suppressed=False):
        d = {"names": [str(i) for i in names],
             "do_approximate_matching": bool(do_approximate_matching),
             "include_suppressed": bool(include_suppressed),
             }
        if context_name is not None:
            d['context_name'] = str(context_name)
        return self._call_api('tnrs/match_names', data=d, demand_success=False)

    def tree_of_life_about(self):
        return self._call_api('tree_of_life/about')

    def tree_of_life_induced_subtree(self, node_ids=None, ott_ids=None, label_format="name_and_id"):
        d = {"label_format": label_format.lower().strip()}
        if not (node_ids or ott_ids):
            raise ValueError("Either node_ids or ott_ids must be provided")
        if node_ids:
            d["node_ids"] = [str(i) for i in node_ids]
        if ott_ids:
            d["ott_ids"] = [int(i) for i in ott_ids]
        return self._call_api('tree_of_life/induced_subtree', data=d, demand_success=False)

    def tree_of_life_mrca(self, node_ids=None, ott_ids=None):
        d = {}
        if not (node_ids or ott_ids):
            raise ValueError("Either node_ids or ott_ids must be provided")
        if node_ids:
            d["node_ids"] = [str(i) for i in node_ids]
        if ott_ids:
            d["ott_ids"] = [int(i) for i in ott_ids]
        return self._call_api('tree_of_life/mrca', data=d, demand_success=False)

    def tree_of_life_node_info(self, node_ids=None, node_id=None, ott_id=None, include_lineage=False):
        d = {"include_lineage": bool(include_lineage)}
        cdict = {"node_ids": node_ids, "node_id": node_id, "ott_id": ott_id}
        name, value = self._one_and_only_one('tree_of_life/node_info', cdict)
        if name == "node_ids":
            d["node_ids"] = [str(i) for i in value]
        elif name == "ott_id":
            d["ott_id"] = int(value)
        else:
            assert name == "node_id"
            d["node_id"] = str(value)
        return self._call_api('tree_of_life/node_info', data=d, demand_success=False)

    def tree_of_life_subtree(self, node_id=None, ott_id=None,
                             tree_format="newick", label_format="name_and_id",
                             height_limit=None):
        d = {"format": str(tree_format), "label_format": str(label_format)}
        if height_limit is not None:
            d["height_limit"] = int(height_limit)
        if node_id is not None:
            d['node_id'] = str(node_id)
        elif ott_id is not None:
            d['ott_id'] = int(ott_id)
        else:
            raise ValueError("Either node_id or ott_id must be given to tree_of_life/subtree")
        return self._call_api('tree_of_life/subtree', data=d, demand_success=False)


def ot_datetime_str_to_object(xdatestr):
    return datetime.strptime(xdatestr, '%Y-%m-%d %H:%M:%S')



#!/usr/bin/env python3
# Implementation file for the calling of HTTP methods and deserializing from JSON.
#   This file is intended as a low-level wrapper that most users would never call directly, but which handles
#   peforming the calls (and if requested) keeping a log of methods called or a curl representation of the
#   calls that were performed.
#

import json
import logging
import sys
import re
from enum import Enum
from .node_reference import SynthNodeReference, OTTaxonRef

import requests


def _escape_dq(s):
    """Lightweight escaping function used in writing curl calls...

    """
    if not isinstance(s, str):
        if isinstance(s, bool):
            return 'true' if s else 'false'
        return s
    if '"' in s:
        ss = s.split('"')
        return '"{}"'.format('\\"'.join(ss))
    return '"{}"'.format(s)


class OTWebServicesError(Exception):
    """This type of error is raised when a web-service call fails for a reason that is
    impossible or difficult to diagnose. The string representation of the error should contain
    some helpful information.
    """

    def __init__(self, message, call_record=None):
        super().__init__(message)
        self.call_record = call_record


class OTClientError(OTWebServicesError):
    """This type of error is raised when the calling code does not make a legitimate request
    based on the Open Tree of Life API's (see https://opentreeoflife.github.io/develop/api).
    """

    def __init__(self, message, call_record=None):
        super().__init__(message, call_record=call_record)


class WebServiceCallRecord(object):
    """Wrapper around a web-service call, returned by WebServiceWrapper methods.

    The main client methods to call are:
      * __bool__ (check if status code was 200)
      * __str__ (explanation of the call status
      * `write_response` (writes call explanation and response, if there was one).
    The most commonly used properties:
      * url: string
      * response: a requests response object
      * status_codeL: None or the HTTP status code as an integer
      * response_dict: None, decoding of a JSON response or {'content' : raw_content} (for non-JSON methods)
    If the API call returns some encoding of a tree, then the `tree` property of the WebServiceCallRecord
        can be used to decode the response.
    """
    def __init__(self, service_wrapper, url, http_method, headers, data):
        self._request_url = url
        self._request_headers = headers
        self._request_http_method = http_method
        self._request_data = data
        self._response_obj = None
        self._response_dict = None
        self._tree = None
        self._node_ref = None
        self._taxon_ref = None
        self._tree_from_response_extractor = None
        try:
            self._to_object_converter = service_wrapper.to_object_converter
        except:
            self._to_object_converter = None

    # noinspection PyPep8
    @property
    def curl_call(self):
        """Returns a string that is a curl representation of the call
        """
        # may want to revisit this implementation

        v = self._request_http_method
        headers = self._request_headers
        data = self._request_data
        url = self._request_url
        varg = '' if v == 'GET' else '-X {} '.format(v)
        if headers:
            hal = ['-H {}:{}'.format(_escape_dq(k), _escape_dq(v)) for k, v in headers.items()]
            hargs = ' '.join(hal)
        else:
            hargs = ''
        dargs = " --data '{}'".format(json.dumps(data)) if data else ''
        return 'curl {v} {h} {u}{d}'.format(v=varg, u=url, h=hargs, d=dargs)

    def __str__(self):
        """Returns and explanation of the URL and status of the call."""
        prefix = "Web-service call to {}".format(self._request_url)
        if self:
            return '{} succeeded.'.format(prefix)
        elif self._response_obj is None:
            return '{} has not been completed (in progress or has not been triggered yet).'.format(prefix)
        return '{} failed with http_status_code={}'.format(prefix, self.status_code)

    def write_response(self, out):
        out.write(str(self))
        if self._response_obj is None:
            out.write('\n')
            return
        out.write(' Response:\n')
        rdict = self.response_dict
        sf = json.dumps(rdict, sort_keys=True, indent=2, separators=(',', ': '), ensure_ascii=True)
        out.write('{}\n'.format(sf))

    @property
    def url(self):
        return self._request_url

    @property
    def response(self):
        return self._response_obj

    @property
    def status_code(self):
        return None if self._response_obj is None else self._response_obj.status_code

    @property
    def response_dict(self):
        if self._response_dict is None:
            if self._response_obj is None:
                return None
            try:
                self._response_dict = self._response_obj.json()  # NOTE: if response is not JSON this will fail
            except json.decoder.JSONDecodeError:
                self._response_dict = {'content': self._response_obj.content}
            # TODO make this not fail on Newick/NEXUS responses
        return self._response_dict

    def __bool__(self):
        """Returns True if call completed with an HTTP status of 200"""
        sc = self.status_code
        return sc is not None and sc == 200

    @property
    def taxon(self):
        if self._taxon_ref is None:
            if not self:
                return None
            self._taxon_ref = OTTaxonRef(self.response_dict)
        return self._taxon_ref

    @property
    def node_ref(self):
        if self._node_ref is None:
            if not self:
                return None
            self._node_ref = SynthNodeReference(self.response_dict)
        return self._node_ref

    @property
    def tree(self):
        if self._tree is None:
            if not self:
                return None
            extractor = self._tree_from_response_extractor
            if extractor is None:
                extractor = default_tree_extractor(self._to_object_converter)
            self._tree = extractor(self.response_dict)
        return self._tree

def extract_content_from_raw_text_method_dict(response_dict):
    return response_dict['content']

def extract_newick(response_dict):
    return response_dict['newick']

def extract_newick_then_obj(response_dict, to_obj_conv):
    newick = extract_newick(response_dict)
    return to_obj_conv.tree_from_newick(newick, suppress_internal_node_taxa=True)

def default_tree_extractor(to_obj_conv):
    if to_obj_conv is None:
        return extract_newick
    return lambda rd: extract_newick_then_obj(rd, to_obj_conv)

class WebServiceRunMode(Enum):
    RUN = 1
    CURL = 2
    CURL_ON_EXIT = 3


class WebServiceWrapper(object):
    def __init__(self, api_endpoint, run_mode=WebServiceRunMode.RUN):
        self._run_mode = run_mode
        self._generate_curl = run_mode in [WebServiceRunMode.CURL, WebServiceRunMode.CURL_ON_EXIT]
        self._perform_ws_calls = run_mode != WebServiceRunMode.CURL
        self._api_endpoint = api_endpoint
        self._api_version = 'v3'
        self._store_responses = False
        self._store_api_calls = True
        self.curl_strings = []
        self.call_history = []
        self.to_object_converter = None

    def _call_api(self, method_url_fragment, data=None,
                  http_method='POST', demand_success=True, headers=None):
        """Returns a ws_call_rec"""
        url = self.make_url(method_url_fragment)
        if headers is None:
            headers = {'content-type': 'application/json', 'accept': 'application/json', }
        elif isinstance(headers, str) and headers.lower() == 'text':
            headers = {'content-type': 'text/plain', 'accept': 'text/plain', }
        try:
            ws_call_rec = self._http_request(url, http_method, data=data, headers=headers)
            if demand_success and not ws_call_rec:
                if not self._perform_ws_calls:
                    return None
                m = 'Wrong HTTP status code from server. Expected 200. Got {}.'
                m = m.format(ws_call_rec.status_code)
                raise OTWebServicesError(m, ws_call_rec)
            return ws_call_rec
        except:
            logging.exception("Error in {} to {}".format(http_method, url))
            raise

    def make_url(self, frag, front_end=False):
        while frag.startswith('/'):
            frag = frag[1:]
        while frag.startswith('/'):
            frag = frag[1:]
        while frag.endswith('/'):
            frag = frag[:-1]
        if self._api_endpoint == 'production':
            if front_end:
                return 'https://tree.opentreeoflife.org/{}/{}'.format(self._api_version, frag)
            return 'https://api.opentreeoflife.org/{}/{}'.format(self._api_version, frag)
        if self._api_endpoint == 'files':
            return 'https://files.opentreeoflife.org/{}'.format(frag)
        if self._api_endpoint == 'dev':
            if front_end:
                return 'https://devtree.opentreeoflife.org/{}/{}'.format(self._api_version, frag)
            return 'https://devapi.opentreeoflife.org/{}/{}'.format(self._api_version, frag)
        if self._api_endpoint == 'next':
            return 'https://nexttree.opentreeoflife.org/{}/{}'.format(self._api_version, frag)
        if self._api_endpoint == 'local':
            tax_pat = re.compile(r'^(v[0-9.]+)/([a-z_]+)/(.+)$')
            m = tax_pat.match(frag)
            if m:
                vers, top_level, tail_frag = m.groups()
                if top_level in ('taxonomy', 'tnrs'):
                    t = 'http://localhost:7474/db/data/ext/{}_{}/graphdb/{}'
                    return t.format(top_level, vers, tail_frag)
                elif top_level in ('tree_of_life',):
                    t = 'http://localhost:6543/{}/{}/{}'
                    return t.format(vers, top_level, tail_frag)
            raise NotImplemented('non-taxonomy local system_to_test')
        if self._api_endpoint.startswith('ot'):
            return 'https://{}.opentreeoflife.org/{}/{}'.format(self._api_endpoint, self._api_version, frag)
        if self._api_endpoint[0].isdigit():
            return 'http://{}/{}/{}'.format(self._api_endpoint, self._api_version, frag)
        raise OTClientError('api_endpoint = "{}" is not supported'.format(self._api_endpoint))

    def _http_request(self, url, http_method="GET", data=None, headers=None):
        """Performs an HTTP call and returns a WebServiceCallRecord instance."""
        rec = WebServiceCallRecord(self, url, http_method, headers, data)
        if self._store_api_calls:
            self.call_history.append(rec)
        if self._generate_curl:
            self.curl_strings.append(rec.curl_call)
        if not self._perform_ws_calls:
            if self._run_mode == WebServiceRunMode.CURL:
                sys.stderr.write('{}\n'.format(self.curl_strings[-1]))
            return rec
        if data:
            resp = requests.request(http_method, url, headers=headers, data=json.dumps(data), allow_redirects=True)
        else:
            resp = requests.request(http_method, url, headers=headers, allow_redirects=True)
        rec._response_obj = resp
        logging.debug('Sent {v} to {s}'.format(v=http_method, s=resp.url))
        return rec



import unittest
from dendropy import Tree
from opentree import OT
from opentree.ws_wrapper import OTWebServicesError


class TestInducedSynth(unittest.TestCase):
    def test_demands_id_arg(self):
        with self.assertRaises(ValueError):
            OT.synth_induced_tree()

    def test_int_ott_ids(self):
        with self.assertRaises(ValueError):
            OT.synth_induced_tree(ott_ids=["hi"])

    def test_success_ott_ids(self):
        OT.synth_induced_tree(ott_ids=[417950, 770315])
    
    def test_success_node_ids(self):
        OT.synth_induced_tree(node_ids=['ott417950', 'ott770315'])
    
    def test_success_mixed_ids(self):
        OT.synth_induced_tree(node_ids=['ott770315'], ott_ids=[417950])

    def test_unknown_ids(self):
        with self.assertRaises(OTWebServicesError):
            OT.synth_induced_tree(node_ids=['ott770315'], ott_ids=[417950, 999999999], ignore_unknown_ids=False)

class TestSynthNodes(object):
    """docstring for TestSynthNodeInfo"""
    def test_synth_node_info(self):
        resp = OT.synth_node_info(node_id='mrcaott354607ott374748')
        assert resp.response_dict['supported_by'] == {'ot_1344@Tr105486': 'Tn16531763'}

    def test_synth_subtree_node(self):
        OT.synth_subtree(node_id='mrcaott354607ott374748')

    def test_synth_subtree_bad_ott(self):
        assert 'broken' in OT.synth_subtree(ott_id='1066581').response_dict


    def test_synth_subtree_good_ott(self):
        res = OT.synth_subtree(ott_id='770309')
        assert isinstance(res.tree, Tree)
        


if __name__ == '__main__':
    unittest.main()



import unittest

from opentree import OT, object_conversion
from dendropy import Tree
DC = object_conversion.DendropyConvert()

study_id = 'ot_1979'
tree_id = 'tree1'

class TestChronoSynthDeps(unittest.TestCase):
    def test_find_chrono(self):
        output = OT.find_trees(search_property="ot:branchLengthMode", value="ot:time")
        chronograms = set()
        for study in output.response_dict["matched_studies"]:
            study_id = study['ot:studyId']
            for tree in study['matched_trees']:
                tree_id = tree['ot:treeId']
                chronograms.add('{}@{}'.format(study_id, tree_id))
        assert 'ot_1000@tree1' in chronograms
        
    def test_dp_convert(self):
        study = OT.get_study(study_id)
        study_nexson = study.response_dict['data']
        tree_obj = DC.tree_from_nexson(study_nexson, tree_id)
        assert isinstance(tree_obj, Tree)
